SystemOrganization addCategory: #TypeFeedback!

Array variableSubclass: #TFMonomorphicCache
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TypeFeedback'!

!TFMonomorphicCache commentStamp: 'mha 6/22/2007 19:41' prior: 0!
An instance of this class represents an "inline cache" object, i.e., an object that caches, for a given send site, the most recently resolved concrete method implementation. The class is a subclass of Array. Its instances are 5 slots long:

	1:	selector			-	the selector sent at this send site (Symbol)
	2:	targetMethod	-	the most recently resolved concrete method implementation (CompiledMethod)
	3:	targetClass		-	the most recently resolved concrete target class (Class)
	4:	hotness			-	the hotness counter for this send (SmallInteger)
	5:	trips			-	the trip counter for this cache (SmallInteger)

The hotness counter is incremented each time the send is executed. The trip counter is incremented each time the cached class and method are replaced with new values because the cache has produced a miss.!

!TFMonomorphicCache class methodsFor: 'instance creation' stamp: 'mha 6/18/2007 21:00'!
defaultSize
	"Return the size of a TFSingleEntryCache object. It contains 5 entries as specified in the class comment."
	
	^ 5! !

!TFMonomorphicCache class methodsFor: 'instance creation' stamp: 'mha 6/22/2007 19:42'!
forSelector: aSelector resolvedClass: aClass resolvedMethod: aCompiledMethod
	"Create an instance of the cache for a given selector and most recently resolved class and method. Set hotness and trip counters to zero."
	
	^ (self new: self defaultSize)
		at: 1 put: aSelector;
		at: 2 put: aCompiledMethod;
		at: 3 put: aClass;
		at: 4 put: 0;		"hotness counter"
		at: 5 put: 0		"trip counter";
		yourself! !

!TFMonomorphicCache class methodsFor: 'resetting' stamp: 'mha 6/21/2007 21:17'!
resetHotnessCounters
	"Reset all hotness counters in monomorphic caches to zero."
	
	self allInstancesDo: [ :mc | mc resetHotnessCounter ]! !

!TFMonomorphicCache class methodsFor: 'resetting' stamp: 'mha 6/21/2007 21:17'!
resetTripCounters
	"Reset all trip counters in monomorphic caches to zero."
	
	self allInstancesDo: [ :mc | mc resetTripCounter ]! !

!TFMonomorphicCache methodsFor: 'accessing' stamp: 'mha 6/21/2007 21:41'!
hotness
	"Answer this monomorphic cache's current hotness counter value."
	
	^ self at: 4! !

!TFMonomorphicCache methodsFor: 'resetting' stamp: 'mha 6/21/2007 21:18'!
resetHotnessCounter
	"Reset this cache's hotness counter to zero."
	
	self at: 4 put: 0! !

!TFMonomorphicCache methodsFor: 'resetting' stamp: 'mha 6/21/2007 21:18'!
resetTripCounter
	"Reset this cache's trip counter to zero."
	
	self at: 5 put: 0! !

!TFMonomorphicCache methodsFor: 'accessing' stamp: 'mha 6/21/2007 21:42'!
selector
	"Answer the selector at this cache's send site."
	
	^ self at: 1! !

!TFMonomorphicCache methodsFor: 'accessing' stamp: 'mha 6/22/2007 19:42'!
targetClass
	"Answer the currently cached target class, i.e., the most recently looked-up one."
	
	^ self at: 3! !

!TFMonomorphicCache methodsFor: 'accessing' stamp: 'mha 6/22/2007 19:42'!
targetMethod
	"Answer the currently cached CompiledMethod, i.e., the most recently looked-up one."
	
	^ self at: 2! !

!TFMonomorphicCache methodsFor: 'accessing' stamp: 'mha 6/21/2007 21:43'!
trips
	"Answer how many times the cached data have been replaced."
	
	^ self at: 5! !

Array variableSubclass: #TFPolymorphicCache
	instanceVariableNames: 'nextFreeEntry'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TypeFeedback'!

!TFPolymorphicCache commentStamp: 'mha 6/18/2007 21:16' prior: 0!
An instance of this class represents a polymorphic cache, i.e., a send site-local cache that stores a given number of concrete target methods for the send in question. The class is a subclass of Array. Its instances are 1+(3*n) slots long, where n is the number of concrete receivers for the send.

At index 1, a polymorphic cache always contains the selector (a Symbol) denoting the message sent at this send site.

All subsequent indices are grouped in triples. Each triple contains, for the cache entry k, at the array index given below, the following data:

	k*3-1	:	the resolved target method (CompiledMethod)
	k*3		:	the concrete resolved target class (Class)
	k*3+1	:	the hotness counter for this particular cache entry (SmallInteger)

The order of concrete resolved method and class is reversed as compared to TFMonomorphicCache to allow for faster class lookup (one less arithmetic operation).

The hotness counter is maintained to allow for later optimisation (e.g., inlining). It is incremented each time the particular cache entry is hit.!

!TFPolymorphicCache class methodsFor: 'instance creation' stamp: 'mha 6/18/2007 21:00'!
defaultSize
	"Answer the number of entries that a newly created polymorphic cache should have."
	
	^ 4! !

!TFPolymorphicCache class methodsFor: 'instance creation' stamp: 'mha 6/21/2007 21:46'!
fromMonomorphicCache: monomorphicCache
	"Create a new polymorphic cache and initialise it with the data found in the monomorphic cache (a TFMonomorphicCache) passed as the argument. The argument's contents are placed in the first entry of the polymorphic cache. All other entries are left blank; they are to be filled subsequently by the virtual machine."
	
	^ (self new: self defaultSize * 3) "one cache entry consumes 3 slots"
		at: 1 put: monomorphicCache selector;
		at: 2 put: monomorphicCache targetClass;
		at: 3 put: monomorphicCache targetMethod;
		at: 4 put: monomorphicCache hotness;
		nextFreeEntry: 2;
		yourself! !

!TFPolymorphicCache class methodsFor: 'resetting' stamp: 'mha 6/21/2007 21:40'!
resetHotnessCounters
	"Reset all polymorphic caches' hotness counters to zero."
	
	self allInstancesDo: [ :pc | pc resetHotnessCounters ]! !

!TFPolymorphicCache methodsFor: 'private' stamp: 'mha 6/21/2007 21:44'!
classIndexForEntry: k
	"Answer the array index at which the cached class for cache entry k is stored."
	
	^ k * 3! !

!TFPolymorphicCache methodsFor: 'accessing' stamp: 'mha 6/21/2007 21:49'!
hotnessAt: k
	"Answer the hotness counter for this cache's k-th entry."
	
	^ self at: (self hotnessIndexForEntry: k)! !

!TFPolymorphicCache methodsFor: 'private' stamp: 'mha 6/21/2007 21:44'!
hotnessAt: k put: hotnessValue
	"Store the given hotness value in the hotness counter for cache entry k."
	
	self at: (self hotnessIndexForEntry: k) put: hotnessValue! !

!TFPolymorphicCache methodsFor: 'private' stamp: 'mha 6/21/2007 21:45'!
hotnessIndexForEntry: k
	"Answer the array index at which the hotness counter for cache entry k is stored."
	
	^ k * 3 + 1! !

!TFPolymorphicCache methodsFor: 'private' stamp: 'mha 6/21/2007 21:45'!
methodIndexForEntry: k
	"Answer the array index at which the cached method for cache entry k is stored."
	
	^ k * 3 - 1! !

!TFPolymorphicCache methodsFor: 'accessing' stamp: 'mha 6/21/2007 21:49'!
nextFreeEntry
	"Answer the number of the next free cache entry."
	
	^ nextFreeEntry! !

!TFPolymorphicCache methodsFor: 'private' stamp: 'mha 6/21/2007 21:45'!
nextFreeEntry: idx
	"Answer the next free cache entry number. Note that this is not the corresponding array index."
	
	nextFreeEntry := idx! !

!TFPolymorphicCache methodsFor: 'resetting' stamp: 'mha 6/21/2007 21:22'!
resetHotnessCounterAt: idx
	"Reset the hotness counters at cache entry idx to zero."
	
	self hotnessAt: idx put: 0! !

!TFPolymorphicCache methodsFor: 'resetting' stamp: 'mha 6/21/2007 21:47'!
resetHotnessCounters
	"Reset all of this cache's hotness counters to zero."
	
	1 to: self nextFreeEntry - 1 do: [ :i | self resetHotnessCounterAt: i ]! !

!TFPolymorphicCache methodsFor: 'accessing' stamp: 'mha 6/21/2007 21:47'!
selector
	"Answer the selector that is sent at this cache's send site."
	
	^ self at: 1! !

!TFPolymorphicCache methodsFor: 'private' stamp: 'mha 6/21/2007 21:48'!
selector: aSymbol
	"Store aSymbol as the selector sent at this cache's send site."
	
	self at: 1 put: aSymbol! !

!TFPolymorphicCache methodsFor: 'accessing' stamp: 'mha 6/21/2007 21:49'!
targetClassAt: k
	"Answer the cached class from this cache's k-th entry."
	
	^ self at: (self classIndexForEntry: k)! !

!TFPolymorphicCache methodsFor: 'accessing' stamp: 'mha 6/21/2007 21:50'!
targetMethodAt: k
	"Answer the cached CompiledMethod from this cache's k-th entry."
	
	^ self at: (self methodIndexForEntry: k)! !
